import numpy as np

from simprotein import run_MC
from optimize_ree import _get_ref_dist

# functions for gradient descent

# compute derivative of Srel with respect to model parameters
def compute_srel_derivative(types, ree_target_dist, bin_edges_target):

    temp = 1.0
    beta = 1. / temp

    # get ratio of distributions
    acc_ratio, ree = run_MC(5000, temp, 2.5, types, save_ree_freq=10)
    ree_model_dist, bin_edges = np.histogram(ree, bins=bin_edges_target, density=True)
    weight_for_srel_deriv = ree_target_dist / (ree_model_dist + 1e-3) # add some epsilon to each bin so that ratio is defined

    # get derivative
    dudxi_weighted, dudxi = run_MC(5000, temp, 2.5, types, srel_deriv_fluc_freq=10, srel_weight=weight_for_srel_deriv, srel_weight_bin_edges=bin_edges_target)
    
    return beta * (dudxi_weighted - dudxi)

# compute derivative of the squared error in the average end-to-end distance
def compute_derivative(types, ree_set):

    temp = 1.0
    beta = 1. / temp

    acc_ratio, ree, alcfluc_mat = run_MC(1000, temp, 2.5, types, save_ree_freq=10, save_alcfluc_freq=10)

    ree_mean = np.mean(ree)
    
    dreedxi = beta / 2. * (alcfluc_mat[:, 2] - alcfluc_mat[:, 0] * alcfluc_mat[:, 1])
    
    return dreedxi * 2 * (ree_mean - ree_set)

# gradient descent step for optimizing squared error in average end-to-end distance
def gradient_descent_step(types, ree_set):

    step_size = 0.1

    grad_f = compute_derivative(types, ree_set)
    types_new = types - grad_f * step_size
    
    types_new = np.maximum(types_new, np.ones(len(types_new)) * -1)
    types_new = np.minimum(types_new, np.ones(len(types_new)))
    
    return types_new

# gradient descent step for optimizing Srel
def gradient_descent_step_srel(types, ree_target_dist, bin_edges_target):

    step_size = 0.1

    grad_f = compute_srel_derivative(types, ree_target_dist, bin_edges_target)
    #print('grad_f', grad_f)
    types_new = types - grad_f * step_size
    
    types_new = np.maximum(types_new, np.ones(len(types_new)) * -1)
    types_new = np.minimum(types_new, np.ones(len(types_new)))
    
    return types_new    

# optimize squared error of mean end-to-end distance
def optimize_ree(ree_set):

    # initialize
    types = np.array([-0.5, -0.5, -0.5, -0.5, -0.5, -0.5, -0.5, -0.5, -0.5, -0.5, -0.5, -0.5, -0.5, -0.5, -0.5])
    n_steps = 100
    for i_step in range(n_steps):
        types = gradient_descent_step(types, ree_set)
        print(types)
    return types

# optimize Srel
def optimize_ree_dist(traj_file=None):

    # initialize
    #types = np.random.random(size=15)*2-1 # start from random sequence
    types = np.array([-0.9, -0.9, -0.9, -0.9, -0.9, -0.9, -0.9, -0.9, -0.9, -0.9, -0.9, -0.9, -0.9, -0.9, -0.9])

    # get target distribution
    ree_target_dist, bin_edges_target = _get_ref_dist(fname_ref_dist='ree_seqn1n1n1n1n1n1n1n1n1n1n1n1n1n1n1_T1.0_run1.txt', bin_edges=np.arange(0, 20.5, 0.5))

    if traj_file is not None:
        f_out = open(traj_file, 'w')

    n_steps = 100

    for i_step in range(n_steps):
        types = gradient_descent_step_srel(types, ree_target_dist, bin_edges_target)
        if traj_file is not None:
            print(types)
            f_out.write(('%6.4f ' * len(types)) % (tuple(types)))
    if traj_file is not None:
        f_out.close()
        
    return types

#print(optimize_ree(5.88))
optimize_ree_dist(traj_file='grad_descent_traj.txt')
