from sklearn.preprocessing import StandardScaler
from sklearn import linear_model
import numpy as np
import matplotlib.pyplot as plt

# fit Srel to a linear combination of engineered features of the sequence

def _get_ref_dist(fname_ref_dist=None, bin_edges=None):
    if fname_ref_dist == None:
        fname = 'ree_seq1111111111_T1.0.txt'
    else:
        fname = fname_ref_dist
    if bin_edges == None:
        bin_edges = np.arange(0, 10.5, 0.5)
    ree = np.loadtxt(fname)
    hist, bin_edges = np.histogram(ree, bins=bin_edges, density=True)
    return hist, bin_edges

def compute_srel(sim_ree_dist, ref_ree_dist, ref_bin_edges):
    srel = np.sum(ref_ree_dist * np.log((ref_ree_dist + 1e-3) / (sim_ree_dist + 1e-3))) # add some epsilon to each bin so that ratio is defined
    return srel

# generate various features of the sequence
def transform(sequences, type='unmodified'):

    if type == 'sum and symmetry':
        return np.hstack(((np.sum((sequences + 1), axis=1) / 2.).reshape(-1, 1), 
                          (np.sum(sequences * sequences[::-1], axis=1)).reshape(-1, 1)))
    
    if type == 'symmetry':
        return (np.sum(sequences * sequences[::-1], axis=1)).reshape(-1, 1)

    if type == 'sum':
        return (np.sum((sequences+1), axis=1) / 2.).reshape(-1, 1)

    if type == 'cluster':
        n_sequences = np.shape(sequences)[0]
        len_sequences = np.shape(sequences)[1]
        cluster_count = np.zeros((n_sequences, len_sequences * 2))
        for i_sequence in range(n_sequences):
            sequence = sequences[i_sequence, :]
            previous_bead = 0
            for i_bead in range(len_sequences):
                if previous_bead == 0:
                    cluster_size = 1
                elif previous_bead == sequence[i_bead]:
                    cluster_size += 1
                else:
                    if previous_bead == 1:
                        cluster_count[i_sequence, cluster_size-1] += 1
                    else:
                        cluster_count[i_sequence, cluster_size-1 + len_sequences] += 1
                    cluster_size = 1
                previous_bead = sequence[i_bead]
            if previous_bead == 1:
                cluster_count[i_sequence, cluster_size-1] += 1
            else:
                cluster_count[i_sequence, cluster_size-1 + len_sequences] += 1
        return cluster_count

    if type == 'rdf':
        n_sequences = np.shape(sequences)[0]
        len_sequences = np.shape(sequences)[1]
        gr = np.zeros((n_sequences, len_sequences))
        for i_sequence in range(n_sequences):
            sequence = sequences[i_sequence, :]
            previous_bead = 0
            for i_bead in range(len_sequences):
                if sequence[i_bead] == -1:
                    continue
                for j_bead in range(len_sequences):
                    if i_bead == j_bead:
                        continue
                    if sequence[j_bead] == -1:
                        continue
                    gr[i_sequence, j_bead - i_bead] += 1
            norm = np.sum((sequence + 1) / 2.) 
            norm = norm * (norm - 1) / (len_sequences - 1) * 2
            gr[i_sequence, :] = gr[i_sequence, :] / norm
        return gr

    if type == 'coarse grain':
        return np.hstack((((np.sum(sequences[:, :3], axis=1) > 0).astype(int).reshape(-1, 1)), 
                          ((np.sum(sequences[:, 3:6], axis=1) > 0).astype(int).reshape(-1, 1)),
                          ((np.sum(sequences[:, 6:9], axis=1) > 0).astype(int).reshape(-1, 1)),
                          ((np.sum(sequences[:, 9:12], axis=1) > 0).astype(int).reshape(-1, 1)),
                          ((np.sum(sequences[:, 12:], axis=1) > 0).astype(int).reshape(-1, 1))))

    if type == 'coarse grain and sums':
        return np.hstack((((np.sum(sequences[:, :3], axis=1) > 0).astype(int).reshape(-1, 1)), 
                          ((np.sum(sequences[:, 3:6], axis=1) > 0).astype(int).reshape(-1, 1)),
                          ((np.sum(sequences[:, 6:9], axis=1) > 0).astype(int).reshape(-1, 1)),
                          ((np.sum(sequences[:, 9:12], axis=1) > 0).astype(int).reshape(-1, 1)),
                          ((np.sum(sequences[:, 12:], axis=1) > 0).astype(int).reshape(-1, 1)),
                          (np.sum((sequences + 1), axis=1) / 2.).reshape(-1, 1),
                          (np.sum(sequences * sequences[::-1], axis=1)).reshape(-1, 1),
                          (np.sum((sequences[:, :4] + sequences[:, -4:]), axis=1).reshape(-1, 1))))

    if type == 'total and ends':
        return np.hstack(((np.sum((sequences + 1), axis=1) / 2.).reshape(-1, 1), 
                          (np.sum((sequences[:, :4] + sequences[:, -4:]), axis=1).reshape(-1, 1))))

    if type == 'unmodified':
        return sequences

def _calc_rmsd(a1, a2):
    da = a2 - a1
    return np.sqrt(np.mean(da * da))

def fit_srel(fname_ref_dist, fname_dataset, conversion_type, plot_label):

    # get reference distribution
    ref_ree_dist, ref_bin_edges = _get_ref_dist(fname_ref_dist)

    # get dataset
    data = np.loadtxt(fname_dataset)
    sequences = data[:, :15]
    pr = data[:, 15:]

    # get features
    seq_transform = transform(sequences, type=conversion_type)

    # compute Srel
    srel = np.array([compute_srel(pr_ind, ref_ree_dist, ref_bin_edges) for pr_ind in pr]).reshape(-1, 1)

    # fit
    lin_reg_model = linear_model.LinearRegression()
    lin_reg_model.fit(seq_transform, srel)

    # compare predicted and actual
    predicted = lin_reg_model.predict(seq_transform)
    plt.plot(srel, predicted, 'o')
    plt.plot([0, 1.25], [0, 1.25])
    plt.text(0.5, 0.1, 'RMSD='+str(round(_calc_rmsd(srel, predicted), 3))+'\n'+plot_label)
    
if __name__ == '__main__':

    ref_ree_dist, ref_bin_edges = _get_ref_dist('ree_seq11111n1n1n1n1n111111_T1.0_run1.txt')

    plt.subplot(2, 2, 1)
    fit_srel('ree_seq11111n1n1n1n1n111111_T1.0.txt', 'pca_data_long.txt', 'unmodified', 'a) sequence')
    plt.xlim([0, 1.25])
    plt.ylim([0, 1.25])
    plt.xticks([], [])
    plt.yticks([0.25, 0.5, 0.75, 1, 1.25], ['0.25', '0.50', '0.75', '1.00', '1.25'])

    plt.subplot(2, 2, 2)
    fit_srel('ree_seq11111n1n1n1n1n111111_T1.0.txt', 'pca_data_long.txt', 'sum', 'b) no. hydro. grps.')
    plt.xlim([0, 1.25])
    plt.ylim([0, 1.25])
    plt.yticks([], [])
    plt.xticks([], [])

    plt.subplot(2, 2, 3)
    fit_srel('ree_seq11111n1n1n1n1n111111_T1.0.txt', 'pca_data_long.txt', 'coarse grain', 'c) max. grps.')
    plt.xlabel(r'Simulated $S_{rel}$')
    plt.ylabel(r'Predicted $S_{rel}$')
    plt.xlim([0, 1.25])
    plt.ylim([0, 1.25])

    plt.subplot(2, 2, 4)
    fit_srel('ree_seq11111n1n1n1n1n111111_T1.0.txt', 'pca_data_long.txt', 'cluster', 'd) cluster dist.')
    plt.xlim([0, 1.25])    
    plt.ylim([0, 1.25])    
    plt.xticks([0.25, 0.5, 0.75, 1, 1.25], ['0.25', '0.50', '0.75', '1.00', '1.25'])
    plt.yticks([], [])

    plt.subplots_adjust(wspace=0, hspace=0)

    plt.savefig('linfits.png', dpi=200, bbox_inches='tight')
