from math import pi, floor
import numpy as np

import proteinlib

import atomwrite

def _norm(r):
    return r / np.sqrt(np.sum(r * r))

def _cos_between(r1, r2):
    return np.sum(r1 * r2) / np.sqrt(np.sum(r1 * r1) * np.sum(r2 * r2))

def generate_trial_segment(r_prev, beta):
    c = beta * 0.25
    while True:
        # generate random vector on unit sphere
        u = np.random.random()
        v = np.random.random()
        theta = 2 * pi * u
        phi = np.arccos(2 * v - 1)
    
        r_new = np.array([np.sin(phi) * np.cos(theta),
                          np.sin(phi) * np.sin(theta),
                          np.cos(phi)])
    
        cos_theta = np.sum(r_new * r_prev) / (np.sqrt(np.sum(r_new * r_new) * np.sum(r_prev * r_prev)))
        u = 0.25 * (1 - cos_theta)
        p_acc = np.exp(-beta * u)
        if p_acc > np.random.random():
            break

    return r_new, u

# k = number of trial segments
# beta = 1 / temperature
# rc = cutoff
# n_regrow = number of regrown segments
# grow_from_end: if True, n_regrow beads are regrown from one end of the polymer, if False, regrown from other end
def config_bias_move(pos, beta, types, rc, k, n_regrow, grow_from_end):

    n_atoms, _ = np.shape(pos)
    n_keep = n_atoms - n_regrow

    if grow_from_end:
        pos_new = np.copy(pos)
        pos_old = np.copy(pos)
        types_test = np.copy(types)
    else:
        pos_new = np.copy(pos[::-1, :])
        pos_old = np.copy(pos[::-1, :])
        types_test = types[::-1]

    # compute change in energy
    du_propose = 0

    # compute Rosenbluth factor of new chain
    Wext = 1
    r_prev = pos_new[n_keep-1, :]-pos_new[n_keep-2, :]
    for i_regrow in range(n_regrow):

        u_ext_segments = np.zeros(k)
        u_bond_segments = np.zeros(k)
        r_new_save = np.zeros((k, 3))
        
        # generate k trial segments
        for i in range(k):
            r_new, u_bond_segments[i] = generate_trial_segment(r_prev, beta)
            r_new_save[i, :] = r_new
            pos_new[n_keep + i_regrow, :] = r_new + pos_new[n_keep + i_regrow - 1, :]
            u_ext_segments[i] = proteinlib.calcextenergysingleparticle(pos_new[:n_keep + i_regrow + 1], 
                                                                   types_test[:n_keep + i_regrow + 1],
                                                                   rc,
                                                                   n_keep + i_regrow)
        # calculate Rosenbluth factor of this segment
        trial_segment_weights = np.exp(-beta * u_ext_segments)
        wext_i = np.sum(trial_segment_weights)
        Wext *= wext_i

        # select one out of the k trial segment weighted by the Boltzmann factors
        i_segment_chosen = np.random.choice(np.arange(k), p=trial_segment_weights/wext_i)
        r_prev = r_new_save[i_segment_chosen, :]
        pos_new[n_keep + i_regrow, :] = r_prev + pos_new[n_keep + i_regrow - 1, :]
        du_propose += u_bond_segments[i_segment_chosen] + u_ext_segments[i_segment_chosen]

    # compute Rosenbluth factor of old chain
    Wext_old = 1
    for i_regrow in range(n_regrow):
        r_prev = pos_old[n_keep+i_regrow-1, :]-pos_old[n_keep+i_regrow-2, :]
        u_ext_segments = np.zeros(k)
        u_bond_segments = np.zeros(k)

        u_ext_segments[k-1] = proteinlib.calcextenergysingleparticle(pos_old[:n_keep + i_regrow + 1, :], 
                                                                     types_test[:n_keep + i_regrow + 1],
                                                                     rc,
                                                                     n_keep + i_regrow)
        for i in range(k - 1):
            r_new, u_bond_segments[i] = generate_trial_segment(r_prev, beta)
            pos_old[n_keep + i_regrow, :] = r_new + pos_old[n_keep + i_regrow - 1, :]
            u_ext_segments[i] = proteinlib.calcextenergysingleparticle(pos_old[:n_keep + i_regrow + 1], 
                                                                   types_test[:n_keep + i_regrow + 1],
                                                                   rc,
                                                                   n_keep + i_regrow)
        if grow_from_end:
            pos_old[n_keep + i_regrow, :] = pos[n_keep + i_regrow, :]
        else:
            pos_old[n_keep + i_regrow, :] = pos[-(n_keep + i_regrow + 1), :]
        u_ext_segments[k-1] = proteinlib.calcextenergysingleparticle(pos_old[:n_keep + i_regrow + 1, :], 
                                                                     types_test[:n_keep + i_regrow + 1],
                                                                     rc,
                                                                     n_keep + i_regrow)
        u_bond_segments[k-1] = 0.25 * (1 - _cos_between(pos_old[n_keep + i_regrow] - pos_old[n_keep + i_regrow - 1],
                                                        pos_old[n_keep + i_regrow - 1] - pos_old[n_keep + i_regrow - 2]))

        trial_segment_weights = np.exp(-beta * u_ext_segments)
        wext_i = np.sum(trial_segment_weights)
        Wext_old *= wext_i
        du_propose -= u_bond_segments[k-1] + u_ext_segments[k-1]
   
    p_acc = min(1, Wext/Wext_old)

    move_accepted = False
    du = 0
    if np.random.random() < p_acc:
        move_accepted = True
        du = du_propose
        if grow_from_end:
            pos = pos_new
        else:
            pos = pos_new[::-1, :]

    return pos, move_accepted, du

# generate labels for atoms according to types for visualization
def _types_to_label(types):
    atom_types = ['C', 'N']
    types_index = ((np.array(types) + 1) / 2).astype(int)
    return [atom_types[type] for type in types_index]

def run_MC(n_steps, T, rc, types, save_ree_freq=-1, 
           save_fluc_freq=-1, 
           save_alcfluc_freq=-1,
           srel_deriv_fluc_freq=-1,
           srel_weight=None,
           srel_weight_bin_edges=None, 
           save_pdb_freq=-1):

    chain_length = len(types)

    # HARD-CODED move parameters
    # regrow half of entire chain
    n_trial_segments = 20
    n_regrow = floor((chain_length-1)/2)

    pos = np.zeros((chain_length, 3))

    # initial position: in a line
    pos[:, 0] = np.arange(chain_length)

    beta = 1./T
    
    n_accepted = 0

    u = proteinlib.calcenergy(pos, types, rc)

    # saving various quantities
    if save_ree_freq != -1:
        ree_sq = np.zeros(floor(n_steps/save_ree_freq))
    if save_fluc_freq != -1:
        fluc_mat = np.zeros(3)
        n_counts = 0
    if save_alcfluc_freq != -1:
        alcfluc_mat = np.zeros((chain_length, 3))
        alcfluc_mat_temp = np.zeros(chain_length)
        n_counts = 0
    if srel_deriv_fluc_freq != -1:
        dudxi_config = np.zeros(chain_length)
        dudxi_weighted = np.zeros(chain_length)
        dudxi = np.zeros(chain_length)
        n_counts_srel = 0

    if save_pdb_freq != -1:
        Pdb = atomwrite.pdbfile('run.pdb', chain_length * 2, Compressed=False, 
                                AtomNames=_types_to_label(types))


    for i_step in range(n_steps):

        # do CBMC step
        if np.random.random() > 0.5:
            pos, move_accepted, du = config_bias_move(pos, beta, types, rc, n_trial_segments, n_regrow, True)
        else:
            pos, move_accepted, du = config_bias_move(pos, beta, types, rc, n_trial_segments, n_regrow, False)

        # track acceptance for acceptance ratio
        if move_accepted:
            n_accepted += 1
            u += du

        # save pdb
        if save_pdb_freq != -1 and i_step % save_pdb_freq == 0:
            Pdb.write(pos)

        # save end-to-end distance
        if save_ree_freq != -1 and i_step % save_ree_freq == 0:
            dr = pos[-1]-pos[0]
            ree_sq[floor(i_step/save_ree_freq)] = np.sum(dr * dr)
            
        # save fluctuations with respect to energy
        if save_fluc_freq != -1 and i_step % save_fluc_freq == 0:
            if save_ree_freq != -1 and i_step % save_ree_freq == 0:
                ree = np.sqrt(ree_sq[floor(i_step/save_ree_freq)])
            else: # calculate ree if we haven't already done so
                dr = pos[-1]-pos[0]
                ree = np.sqrt(np.sum(dr * dr))
            fluc_mat[0] += u
            fluc_mat[1] += ree
            fluc_mat[2] += u * ree
            n_counts += 1

        # save fluctuations with respect to derivative of energy with respect to type parameter
        if save_alcfluc_freq != -1 and i_step % save_alcfluc_freq == 0:
            if save_ree_freq != -1 and i_step % save_ree_freq == 0:
                ree = np.sqrt(ree_sq[floor(i_step/save_ree_freq)])
            else: # calculate ree if we haven't already done so
                dr = pos[-1]-pos[0]
                ree = np.sqrt(np.sum(dr * dr))
            alcfluc_mat_temp = proteinlib.calcalchemicalfluctuations(pos, types, rc, alcfluc_mat_temp)
            alcfluc_mat[:, 0] += alcfluc_mat_temp
            alcfluc_mat[:, 1] += ree
            alcfluc_mat[:, 2] += alcfluc_mat_temp * ree
            n_counts += 1

        # do full energy update
        if i_step % 1000 == 0:
            u = proteinlib.calcenergy(pos, types, rc)
            
        # save derivatives for computing derivative of Srel
        if srel_deriv_fluc_freq != -1 and i_step % srel_deriv_fluc_freq == 0:
            dr = pos[-1]-pos[0]
            ree = np.sqrt(np.sum(dr * dr))
            srel_weight_ree = srel_weight[np.digitize(ree, srel_weight_bin_edges)-1]
            dudxi_config = proteinlib.calcalchemicalfluctuations(pos, types, rc, dudxi_config)
            dudxi_weighted += dudxi_config * srel_weight_ree
            dudxi += dudxi_config
            n_counts_srel += 1

    if save_pdb_freq != -1:
        Pdb.close()

    if srel_deriv_fluc_freq != -1:
        return -0.5*dudxi_weighted/n_counts_srel, -0.5*dudxi/n_counts_srel
    if save_ree_freq != -1:
        if save_fluc_freq != -1:
            return float(n_accepted)/float(n_steps), np.sqrt(ree_sq), fluc_mat/n_counts
        if save_alcfluc_freq != -1:
            return float(n_accepted)/float(n_steps), np.sqrt(ree_sq), alcfluc_mat/n_counts
        return float(n_accepted)/float(n_steps), np.sqrt(ree_sq)
    if save_fluc_freq != -1:
        return float(n_accepted)/float(n_steps), fluc_mat/n_counts
    if save_alcfluc_freq != -1:
        return float(n_accepted)/float(n_steps), alcfluc_mat/n_counts
    return float(n_accepted)/float(n_steps)

def _inttostring(a):
    if a < 0:
        return 'n' + str(-a)
    else:
        return str(a)

if __name__ == '__main__':        
    temp = 1.0
    types = [1, 1, -1, -1, 1, -1, -1, -1, -1, 1, -1, 1, 1, 1, 1]

    acc_ratio, ree, alcfluc_mat = run_MC(10000, temp, 2.5, types, save_ree_freq=10, save_alcfluc_freq=10, save_pdb_freq=10)

    print('acceptance ratio', acc_ratio)

    type_str = ''.join([_inttostring(type) for type in types])

    np.savetxt('ree_seq' + type_str + '_T' + str(temp) + '.txt', ree)
    np.savetxt('alcfluc_mat' + type_str + '_T' + str(temp) + '.txt', alcfluc_mat)
